/*============================================================================
  File:     Columnstore & Partitioning.sql

  Summary:  This script was created as part of SQL Server 2012 Early Adoption Cook Book
            http://social.technet.microsoft.com/wiki/contents/articles/6967.sql-server-2012-early-adoption-cook-book.aspx

            Accelerating Data Warehouse Queries with SQL Server 2012 Columnstore Indexes

            NOTE: Demo was written using 4 logical processors and parallelism 
            enabled (in order to show batch processing)

  SQL Server Versions: SQL Server 2012
------------------------------------------------------------------------------
  Written by SQLskills.com

  (c) SQLskills.com. All rights reserved.

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  You may alter this code for your own *non-commercial* purposes. You may
  republish altered code as long as you include this copyright and give due
  credit, but you must obtain prior permission before blogging this code.
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

USE AdventureWorksDWDenali;
GO

-- Let's view CPU, Elapsed time, I/O and the actual execution plan
SET STATISTICS TIME ON;
SET STATISTICS IO ON;

-- Execute this query twice, and measure the second attempt (get past data caching overhead)
SELECT  p.EnglishProductName,
		d.WeekNumberOfYear,
		d.CalendarYear,
		AVG(fpi.UnitCost),
		SUM(fpi.UnitsOut)
FROM	dbo.FactProductInventory as fpi
INNER JOIN dbo.DimProduct as p ON
	fpi.ProductKey = p.ProductKey
INNER JOIN dbo.DimDate as d ON
	fpi.DateKey = d.FullDateAlternateKey
GROUP BY p.EnglishProductName,
		d.WeekNumberOfYear,
		d.CalendarYear
ORDER BY p.EnglishProductName,
		 d.CalendarYear,
		 d.WeekNumberOfYear;

-- What did you see?

SET STATISTICS TIME OFF;
SET STATISTICS IO OFF;

-- From my demo:
-- Nonclustered index: CPU time = 16562 ms,  elapsed time = 5532 ms.
--  Columnstore index: CPU time =  2579 ms,  elapsed time = 1750 ms.

-- Now we'll create the columnstore index
CREATE NONCLUSTERED COLUMNSTORE INDEX [IX_CS_FactProductInventory]
ON dbo.FactProductInventory
( 
		ProductKey, 
		DateKey, 
		UnitCost, 
		UnitsIn, 
		UnitsOut, 
		UnitsBalance
);

-- Let's view CPU, Elapsed time, I/O and the actual execution plan
SET STATISTICS TIME ON;
SET STATISTICS IO ON;

-- Execute this query twice, and measure the second attempt (get past data caching overhead)
SELECT  p.EnglishProductName,
		d.WeekNumberOfYear,
		d.CalendarYear,
		AVG(fpi.UnitCost),
		SUM(fpi.UnitsOut)
FROM	dbo.FactProductInventory as fpi
INNER JOIN dbo.DimProduct as p ON
	fpi.ProductKey = p.ProductKey
INNER JOIN dbo.DimDate as d ON
	fpi.DateKey = d.FullDateAlternateKey
GROUP BY p.EnglishProductName,
		d.WeekNumberOfYear,
		d.CalendarYear
ORDER BY p.EnglishProductName,
		 d.CalendarYear,
		 d.WeekNumberOfYear;

-- What did you see?

SET STATISTICS TIME OFF;
SET STATISTICS IO OFF;

-- Now let's demonstrate updating a CS table with partitioning
-- Removing CS index for now and supporting NCI created earlier
DROP INDEX [IX_CS_FactProductInventory] ON [dbo].[FactProductInventory];
GO

DROP INDEX [NCI_FactProductInventory_DateKey] ON [dbo].[FactProductInventory];
GO

-- Create partition function and scheme
CREATE PARTITION FUNCTION [pfn_DateKey](date) AS RANGE RIGHT 
FOR VALUES (N'2000-01-13', N'2000-02-13', N'2000-03-13', N'2000-04-13', N'2000-05-13', N'2000-06-13', N'2000-07-13',
 N'2000-08-13', N'2000-09-13', N'2000-10-13', N'2000-11-13', N'2000-12-13', N'2001-01-13', N'2001-02-13',
  N'2001-03-13', N'2001-04-13', N'2001-05-13', N'2001-06-13', N'2001-07-13', N'2001-08-13', N'2001-09-13',
   N'2001-10-13', N'2001-11-13', N'2001-12-13', N'2002-01-13', N'2002-02-13', N'2002-03-13', N'2002-04-13', 
   N'2002-05-13', N'2002-06-13', N'2002-07-13', N'2002-08-13', N'2002-09-13', N'2002-10-13', N'2002-11-13', 
   N'2002-12-13', N'2003-01-13', N'2003-02-13', N'2003-03-13', N'2003-04-13', N'2003-05-13', N'2003-06-13',
    N'2003-07-13', N'2003-08-13', N'2003-09-13', N'2003-10-13', N'2003-11-13', N'2003-12-13', N'2004-01-13', 
	N'2004-02-13', N'2004-03-13', N'2004-04-13', N'2004-05-13', N'2004-06-13', N'2004-07-13', N'2004-08-13',
	 N'2004-09-13', N'2004-10-13', N'2004-11-13', N'2004-12-13', N'2005-01-13');
GO

CREATE PARTITION SCHEME [pfn_DateKey] AS PARTITION [pfn_DateKey] 
ALL TO([PRIMARY]);
GO

CREATE CLUSTERED INDEX [CI_DateKey] 
ON [dbo].[FactProductInventory]
	([DateKey], [ProductKey]) ON [pfn_DateKey]([DateKey]);
GO

CREATE NONCLUSTERED INDEX [NCI_UnitCost] 
ON [dbo].[FactProductInventory]
	([UnitCost] ASC)
 ON [pfn_DateKey]([DateKey]);
GO

-- Add in the CS index (notice no need to specify the partition scheme)
CREATE NONCLUSTERED COLUMNSTORE INDEX [NSI_FactProductInventory] 
ON [dbo].[FactProductInventory]
(
	[ProductKey],
	[DateKey],
	[UnitCost],
	[UnitsIn],
	[UnitsOut],
	[UnitsBalance]
);
GO

-- Demo that updates still aren't possible
INSERT dbo.FactProductInventory
(ProductKey, DateKey, UnitCost, UnitsIn, UnitsOut, UnitsBalance)
VALUES (167, '2000-01-14', .28, 0, 0, 870);
GO

-- Now we'll create a new archive table
CREATE TABLE [dbo].[Archive_FactProductInventory](
	[ProductKey] [int] NOT NULL,
	[UnitCost] [money] NOT NULL,
	[UnitsIn] [int] NOT NULL,
	[UnitsOut] [int] NOT NULL,
	[UnitsBalance] [int] NOT NULL,
	[DateKey] [date] NOT NULL
) ON [PRIMARY];
GO

CREATE CLUSTERED INDEX [Archive_FactProductInventory_CI_DateKey] ON [dbo].[Archive_FactProductInventory]
(
	[DateKey] ASC,
	[ProductKey] ASC
)ON [PRIMARY];
GO

CREATE NONCLUSTERED INDEX [Archive_FactProductInventory_NCI_UnitCost] ON [dbo].[Archive_FactProductInventory]
([UnitCost] ASC)
INCLUDE ( 	[DateKey])  ON [PRIMARY];
GO

ALTER TABLE [dbo].[Archive_FactProductInventory]  
WITH CHECK ADD  CONSTRAINT [Archive_FactProductInventory_FK_ProductInventory_DimProduct] 
FOREIGN KEY([ProductKey])
REFERENCES [dbo].[DimProduct] ([ProductKey]);
GO

ALTER TABLE [dbo].[Archive_FactProductInventory] 
 WITH CHECK ADD  CONSTRAINT [chk_Archive_FactProductInventory]
  CHECK  ([DateKey]>=N'2001-06-13' AND [DateKey]<N'2001-07-13');
  GO

-- 6,210,288 rows
SELECT COUNT(*) FROM [AdventureWorksDWDenali].[dbo].[FactProductInventory];

-- 0 rows
SELECT COUNT(*) FROM [AdventureWorksDWDenali].[dbo].[Archive_FactProductInventory];

-- Switch out a partition
ALTER TABLE [AdventureWorksDWDenali].[dbo].[FactProductInventory]
 SWITCH PARTITION 19 TO [AdventureWorksDWDenali].[dbo].[Archive_FactProductInventory];
GO

-- Validate that rows were moved to the archive table
-- 63024 now in the archive table
SELECT COUNT(*)
FROM [AdventureWorksDWDenali].[dbo].[Archive_FactProductInventory];
GO

-- Validate that this happened while normal updates (not partitioned ones, are disallowed)
INSERT dbo.FactProductInventory
(ProductKey, DateKey, UnitCost, UnitsIn, UnitsOut, UnitsBalance)
VALUES (167, '2000-01-14', .28, 0, 0, 870);
GO

-- We can also switch in back data to the main table

-- Should be 6,147,264
SELECT COUNT(*) FROM [dbo].[FactProductInventory];
GO

-- Let's create a CS index on the archive table (you'll see an error if you do not)
CREATE NONCLUSTERED COLUMNSTORE INDEX [NSI_ArchiveFactProductInventory] 
ON [dbo].[Archive_FactProductInventory]
(
	[ProductKey],
	[DateKey],
	[UnitCost],
	[UnitsIn],
	[UnitsOut],
	[UnitsBalance]
)WITH (DROP_EXISTING = OFF);
GO

ALTER TABLE [dbo].[Archive_FactProductInventory] 
SWITCH  TO [dbo].[FactProductInventory]
PARTITION 19;
GO

-- Now check the rowcounts of both tables (it worked!)
-- back to 6,210,288
SELECT COUNT(*) FROM [dbo].[FactProductInventory];
GO








